/*
 * Newsfeeds file entries should look like this:
 *     host.name.of.site[/exclude,exclude,...]\
 *             :pattern,pattern...[/distribution,distribution...]\
 *             :Tf,Wnm
 *             :
 * or
 *     sitename[/exclude,exclude,...]\
 *             :pattern,pattern...[/distribution,distribution...]\
 *             :Tf,Wnm
 *             :host.name.of.site
 *
 * Four files full of
 *    token messageid
 * or might be blanked out
 *    <spc><spc><spc><spc>....
 *
 *   site.name_duct.lock       lock preventing multiple ducts
 *                                holder of this lock is "duct"
 * F site.name                 main feed file
 *                                opened/created, then written, by innd
 *                                read by duct
 *                                unlinked by duct
 *                                tokens blanked out by duct when processed
 * D site.name_duct            temporary feed file during flush (or crash)
 *                                hardlink created by duct
 *                                unlinked by duct
 *   site.name_duct.defer      431'd articles, still being written,
 *                                created, written, used by duct
 *   site.name_backlog.lock    lock taken out by innxmit wrapper
 *                                holder and its child are "xmit"
 *   site.name_backlog_<date>.<inum>
 *                             431'd articles, ready for innxmit
 *                                created (link/mv) by duct
 *                                read by xmit
 *                                unlinked by xmit
 *   site.name_backlog_<letters> eg
 *   site.name_backlog_manual
 *                             anything the sysadmin likes (eg, feed files
 *                             from old feeds to be merged into this one)
 *                                created (link/mv) by admin
 *                                read by xmit
 *                                unlinked by xmit


   OVERALL STATES:

  	    					                START
                                                                  |
                                                             check D, F
                                                                  |
                          <--------------------------------------'|
        Nothing                            F, D both ENOENT       |
         F: ENOENT                             	                  |
         D: ENOENT     	                       	                  |
         duct: not not reading anything                           |
           |                                 	                  |
           |`---------------------.                               |
           |                      | duct times out waiting for F  |
           V  innd creates F      | duct exits                    |
           |                      V                               |
        Noduct                    GO TO Dropped                   |
         F: innd writing                     	                  |
         D: ENOENT                           	                  |
         duct: not running or not reading anything                |
           |                                 	                  |
           |                                 	                  |
     ,-->--+                   <---------------------------------'|
     |     |  duct opens F                         F exists       |
     |     |                                   	   D ENOENT       |
     |     V                            	                  |
     |  Normal                                                    |
     |   F: innd writing, duct reading                            |
     |   D: ENOENT                                                |
     |     |                                                      |
     |     |  duct decides time to flush                          |
     |     |  duct makes hardlink                                 |
     |     |                                                      |
     |     V                            <------------------------'|
     |  Hardlinked                                  F==D          |
     |   F == D: innd writing, duct reading         both exist    |
     ^     |                                                      |
     |     |  duct unlinks F                                   	  |
     |     V                                                  	  |
     |  Moved                               <----+------------<--'|
     |   F: ENOENT                               |  F ENOENT      |
     |   D: innd writing, duct reading           |  D exists      |
     |     |                                     |            	  |
     |     |  duct requests flush of feed        |            	  |
     |     |   (others can too, harmlessly)      |            	  |
     |     V				 	 |	      	  |
     |  Flushing                              	 |	      	  |
     |   F: ENOENT                            	 |	      	  |
     |   D: innd flushing, duct reading        	 |	      	  |
     |     |   	      			       	 |	      	  |
     |     |   inndcomm flush fails    	       	 |	      	  |
     |     |`-------------------------->---------'	      	  |
     |     |   	       			 	 	      	  |
     |     |   inndcomm reports no such site                   	  |
     |     |`---------------------------------------------------- | -.
     |     |	      			 	    	      	  |  |
     |     |  innd finishes writing D, creates F    	      	  |  |
     |     |  inndcomm reports flush successful	    	      	  |  |
     |     |                                                   	  |  |
     |     V                                     	       	  |  |
     |  Separated                                <----------------'  |
     |   F: innd writing                            F!=D             /
     |   D: duct reading                             both exist     /
     |     |                                                       /
     |     |  duct gets to the end of D                           /
     |     |  duct opens F too                                 	 /
     |     V                                                    /
     |  Finishing                                              /
     |   F: innd writing, duct reading                        |
     |   D: duct finishing                                    V
     |     |                                           	Dropping
     |     |  duct finishes processing D       	 	 F: ENOENT
     |     V  duct unlinks D                   	       	 D: duct reading
     |     |                            		      |
     `--<--'                            		      | duct finishes
                                               	       	      |  processing D
                                                              | duct unlinks D
                                                              | duct exits
                                                              V
       	       	       	       	       	       	       	Dropped
  							 F: ENOENT
  							 D: ENOENT
  							 duct not running

   "duct reading" means innduct is reading the file but also
   overwriting processed tokens.

 *
 *
 */


/*----- general definitions, probably best not changed -----*/

#define PERIOD_SECONDS 60

#define CONNCHILD_ESTATUS_STREAM   4
#define CONNCHILD_ESTATUS_NOSTREAM 5


/*----- configuration options -----*/

static char *feedfile;
static int max_connections, max_queue_per_conn;
static int connection_setup_timeout, port, try_stream;
static const char *remote_host;

static double accept_proportion;
static double nocheck_thresh= 0.95;
static double nocheck_decay= 1-1/100;
static int nocheck, nocheck_reported;


/*----- doubly linked lists -----*/

#define ISNODE(T)    T *next, *back;
#define LIST(T)      struct { T *head, *tail, *tailpred; int count; }

#define NODE(n) ((struct node*)&(n)->head)

#define LIST_ADDHEAD(l,n)						\
 (list_addhead((struct list*)&(l), NODE((n))), (void)(l).count++)
#define LIST_ADDTAIL(l,n)						\
 (list_addtail((struct list*)&(l), NODE((n))), (void)(l).count++)

#define LIST_REMHEAD(l)							  \
 ((l).count ? ((l).count--, (void*)list_remhead((struct list*)&(l))) : 0)
#define LIST_REMTAIL(l)							  \
 ((l).count ? ((l).count--, (void*)list_remtail((struct list*)&(l))) : 0)
#define LIST_REMOVE(l,n)			\
 (list_remove(NODE((n))), (void)(l).count--)
#define LIST_INSERT(l,n,pred) \
 (list_insert((struct list*)&(l), NODE((n)), NODE((pred))), (void)(l).count++)


/*----- statistics -----*/

#define RESULT_COUNTS				\
  RC(offered)					\
  RC(sent)					\
  RC(unwanted)					\
  RC(accepted)					\
  RC(rejected)					\
  RC(deferred)

typedef enum {
#define RC_INDEX(x) RCI_##x
  RESULT_COUNTS
  RCI_max
} ResultCountIndex;

typedef struct {
  int articles[2 /* checked */][RCI_max];
} Counts;


/*----- transmission buffers -----*/

#define CONNIOVS 128

typedef enum {
  xk_Malloc, xk_Const, xk_Artdata;
} XmitKind;

typedef struct {
  XmitKind kind;
  union {
    char *malloc_tofree;
    ARTHANDLE *sm_art;
  } info;
} XmitDetails;


/*----- core operational data structure types -----*/

struct Article {
  int midlen;
  int checked, sentbody;
  InputFile *ipf;
  TOKEN token;
  off_t offset;
  int blanklen;
  char messageid[1];
};

typedef struct {
  /* This is an instance of struct oop_readable */
  struct oop_readable readable; /* first */
  oop_readable_call *readable_callback;
  void *readable_callback_user;

  int fd;
  const char *path; /* ptr copy of path_<foo> or feedfile */
  struct Filemon_Perfile *filemon;

  oop_read *rd;
  long inprogress; /* no. of articles read but not processed */
  off_t offset;
} InputFile;

typedef enum {
  sm_WAITING,
  sm_NORMAL,
  sm_FLUSHING,
  sm_FLUSHFAIL,
  sm_DROPPING,
  sm_SEPARATED,
  sm_FINISHING;
} StateMachineState;

struct Conn {
  ISNODE(Conn);
  int fd, max_queue, stream;
  LIST(Article) queue; /* not yet told peer, or CHECK said send it */
  LIST(Article) sent; /* offered/transmitted - in xmit or waiting reply */
  struct iovec xmit[CONNIOVS];
  XmitDetails xmitd[CONNIOVS];
  int xmitu;
};


/*----- operational variables -----*/

static int since_connect_attempt;
static int nconns;
static LIST(Conn) idle, working, full;
static LIST(Article) *queue;

static char *path_ductlock, *path_duct, *path_ductdefer;

static StateMachineState sms;
static FILE *defer;
static InputFile *main_input_file, *old_input_file;
static int waiting_periods_sofar;


/*----- function predeclarations -----*/

static void conn_check_work(Conn *conn);

static int filemon_init(void);
static void filemon_setfile(int mainfeed_fd, const char *mainfeed_path);
static void filemon_callback(void);


/*========== utility functions etc. ==========*/

static void perhaps_close(int *fd) { if (*fd) { close(*fd); fd=0; } }


/*========== making new connections ==========*/

static int connecting_sockets[2]= {-1,-1};
static pid_t connecting_child;

static void report_child_status(const char *what, int status) {
  if (WIFEXITED(status)) {
    int es= WEXITSTATUS(status);
    if (es)
      warn("%s: child died with error exit status %d",es);
  } else if (WIFSIGNALED(status)) {
    int sig= WTERMSIG(status);
    const char *sigstr= strsignal(sig);
    const char *coredump= WCOREDUMP(status) ? " (core dumped)" : "";
    if (sigstr)
      warn("%s: child died due to fatal signal %s%s", what, sigstr, coredump);
    else
      warn("%s: child died due to unknown fatal signal %d%s",
	   what, sig, coredump);
  } else {
    warn("%s: child died with unknown wait status %d", status);
  }
}

static void connect_attempt_discard(void) {
  if (connecting_sockets[0]) {
    cancel_fd(loop, connecting_sockets[0], OOP_READ);
    cancel_fd(loop, connecting_sockets[0], OOP_EXCEPTION);
  }
  perhaps_close(&connecting_sockets[0]);
  perhaps_close(&connecting_sockets[1]);

  if (connecting_child) {
    int status;
    r= kill(connecting_child, SIGKILL);
    if (r) sysdie("cannot kill connect child");

    pid_t got= waitpid(connecting_child, &status, WNOHANG);
    if (got==-1) sysdie("cannot reap connect child");

    if (!(WIFEXITED(status) ||
	  (WIFSIGNALED(status) && WTERMSIG(status) == SIGKILL))) {
      report_child_status("connect"
    }
    connecting_child= 0;
  }
}

#define PREP_DECL_MSG_CMSG(msg)			\
  struct msghdr msg;				\
  memset(&msg,0,sizeof(msg));			\
  char msg##cbuf[CMSG_SPACE(sizeof(fd))];	\
  msg.msg_control= msg##cbuf;			\
  msg.msg_controllen= sizeof(msg##cbuf);

static void *connchild_event(oop_source *lp, int fd, oop_event e, void *u) {
  Conn *conn= 0;

  conn= xcalloc(sizeof(*conn));

  DECL_MSG_CMSG(msg);
  struct cmsghdr *h= 0;
  ssize_t rs= recvmsg(fd, &msg, MSG_DONTWAIT);
  if (rs >= 0) h= CMSG_FIRSTHDR(&msg);
  if (!h) {
    int status;
    pid_t got= waitpid(connecting_child, &status, WNOHANG);
    if (got != -1) {
      assert(got==connecting_child);
      connecting_child= 0;
      if (WIFEXITED(status) &&
	  (WEXITSTATUS(status) != 0
	   WEXITSTATUS(status) != CONNCHILD_ESTATUS_STREAM &&
	   WEXITSTATUS(status) != CONNCHILD_ESTATUS_NOSTREAM)) {
	/* child already reported the problem */
      } else if (WIFSIGNALED(status) && WTERMSIG(status) == SIGALARM) {
	warn("connect: connection attempt timed out");
      } else if (!WIFEXITED(status)) {
	report_child_status("connect", status);
	/* that's probably the root cause then */
      }
    } else {
      /* child is still running apparently, report the socket problem */
      if (rs < 0)
	syswarn("connect: read from child socket failed");
      else if (e == OOP_EXCEPTIONN)
	warn("connect: unexpected exception on child socket");
      else
	warn("connect: unexpected EOF on child socket");
    }
    goto x;
  }

#define CHK(field, val)							  \
  if (h->cmsg_##field != val) {						  \
    die("connect: child sent cmsg with cmsg_" #field "=%d, expected %d"); \
    goto x;								  \
  }
  CHK(level, SOL_SOCKET);
  CHK(type,  SCM_RIGHTS);
  CHK(len,   CMSG_LEN(sizeof(conn-b>fd)));
#undef CHK

  if (CMSG_NXTHDR,&msg,h) { die("connect: child sent many cmsgs"); goto x; }

  memcpy(&conn->fd, CMSG_DATA(h), sizeof(conn->fd));

  pid_t got= waitpid(connecting_child, &status, 0);
  if (got==-1) sysdie("connect: real wait for child");
  assert(got == connecting_child);
  connecting_child= 0;

  if (!WIFEXITED(status)) { report_child_status("connect",status); goto x; }
  int es= WEXITSTATUS(status);
  switch (es) {
  case CONNCHILD_ESTATUS_STREAM:    conn->stream= 1;   break;
  case CONNCHILD_ESTATUS_NOSTREAM:  conn->stream= 0;   break;
  default:
    die("connect: child gave unexpected exit status %d", es);
  }

  set nonblocking;

  /* Phew! */
  LIST_ADDHEAD(idle, conn);
  notice(CN "connected %s", conn->fd, conn->stream ? "streaming" : "plain");
  connect_attempt_discard();
  check_master_queue();
  return 0;

 x:
  if (conn) {
    perhaps_close(&conn->fd);
    free(conn);
  }
  connect_attempt_discard();
}

static void connect_start() {
  assert(!connecting_sockets[0]);
  assert(!connecting_sockets[1]);
  assert(!connecting_child);

  notice("starting connection attempt");

  r= socketpair(AF_UNIX, SOCK_STREAM, 0, connecting_sockets);
  if (r) { syswarn("connect: cannot create socketpair for child"); goto x; }

  connecting_child= fork();
  if (connecting_child==-1) { syswarn("connect: cannot fork"); goto x; }

  if (!connecting_child) {
    FILE *cn_from, *cn_to;
    char buf[NNTP_STRLEN+100];
    int exitstatus= CONNCHILD_ESTATUS_NOSTREAM;

    postfork();

    r= close(connecting_sockets[0]);
    if (r) sysdie("connect: close parent socket in child");

    alarm(connection_setup_timeout);
    if (NNTPconnect(remote_host, port, &cn_from, &cn_to, buf) < 0) {
      if (buf[0]) {
	sanitise_inplace(buf);
	die("connect: rejected: %s", buf);
      } else {
	sysdie("connect: connection attempt failed");
      }
    }
    if (NNTPsendpassword(remote_host, cn_from, cn_to) < 0)
      sysdie("connect: authentication failed");
    if (try_stream) {
      if (fputs("MODE STREAM\r\n", cn_to) ||
	  fflush(cn_to))
	sysdie("connect: could not send MODE STREAM");
      buf[sizeof(buf)-1]= 0;
      if (!fgets(buf, sizeof(buf)-1, cn_from)) {
	if (ferror(cn_from))
	  sysdie("connect: could not read response to MODE STREAM");
	else
	  die("connect: connection close in response to MODE STREAM");
      }
      int l= strlen(buf);
      assert(l>=1);
      if (buf[-1]!='\n') {
	sanitise_inplace(buf);
	die("connect: response to MODE STREAM is too long: %.100s...",
	    remote_host, buf);
      }
      l--;  if (l>0 && buf[1-]=='\r') l--;
      buf[l]= 0;
      char *ep;
      int rcode= strtoul(buf,&ep,10);
      if (ep != buf[3]) {
	sanitise_inplace(buf);
	die("connect: bad response to MODE STREAM: %.50s", buf);
      }
      switch (rcode) {
      case 203:
	exitstatus= CONNCHILD_ESTATUS_STREAM;
	break;
      case 480:
      case 500:
	break;
      default:
	sanitise_inplace(buf);
	warn("connect: unexpected response to MODE STREAM: %.50s", buf);
	exitstatus= 2;
	break;
      }
    }
    int fd= fileno(cn_from);

    PREP_DECL_MSG_CMSG(msg);
    struct cmsghdr *cmsg= CMSG_FIRSTHDR(&msg);
    cmsg->cmsg_level= SOL_SOCKET;
    cmsg->cmsg_type=  SCM_RIGHTS;
    cmsg->cmsg_len=   CMSG_LEN(sizeof(fd));
    memcpy(CMSG_DATA(cmsg), &fd, sizeof(fd));

    msg.msg_controllen= cmsg->cmsg_len;
    r= sendmsg(connecting_sockets[1], &msg, 0);
    if (r) sysdie("sendmsg failed for new connection");

    _exit(exitstatus);
  }

  r= close(connecting_sockets[1]);  connecting_sockets[1]= 0;
  if (r) sysdie("connect: close child socket in parent");

  loop->on_fd(loop, connecting_sockets[0], OOP_READ,      connchild_event, 0);
  loop->on_fd(loop, connecting_sockets[0], OOP_EXCEPTION, connchild_event, 0);
  return OOP_CONTINUE;

 x:
  connect_attempt_discard();
}


/*========== overall control of article flow ==========*/

static void check_master_queue(void) {
  try reading current feed file;

  if (!queue.count)
    return;

  Conn *last_assigned=0;
  for (;;) {
    if (working.head) {
      conn_assign_one_article(&working, &last_assigned);
    } else if (idle.head) {
      conn_assign_one_article(&idle, &last_assigned);
    } else if (nconns < maxconns && queue.count >= max_queue_per_conn &&
	       !connecting_child && !connect_delay) {
      connect_delay= reconnect_delay_periods;
      connect_start();
    } else {
      break;
    }
  }
  conn_check_work(last_assigned);
}

static void conn_assign_one_article(LIST(Conn) *connlist,
				    Conn **last_assigned) {
  Conn *conn= connlist->head;

  LIST_REMOVE(*connlist, conn);
  Article *art= LIST_REMHEAD(queue);
  LIST_ADDTAIL(conn->queue, art);
  LIST_ADD(*conn_determine_right_list(conn), conn);

  /* This slightly odd arrangement is so that we call conn_check_work
   * once after filling the queue for a new connection in
   * check_master_queue, rather than for each article. */
  if (conn != *last_assigned && *last_assigned)
    conn_check_work(*last_assigned);
  *last_assigned= conn;
}

static int conn_total_queued_articles(Conn *conn) {
  return conn->sent.count + conn->queue.count;
}

static LIST(Conn) *conn_determine_right_list(Conn *conn) {
  int inqueue= conn_total_queued_articles(conn);
  assert(inqueue <= max_queue);
  if (inqueue == 0) return &idle;
  if (inqueue == conn->max_queue) return &full;
  return &working;
}

static void *conn_writeable(oop_source *l, int fd, int ev, void *u) {
  check_conn_work(u);
  return OOP_CONTINUE;
}

static void conn_check_work(Conn *conn)  {
  void *rp= 0;
  for (;;) {
    conn_make_some_xmits(conn);
    if (!conn->xmitu) {
      loop->cancel_fd(loop, conn->fd, OOP_WRITE);
      return;
    }

    void *rp= conn_write_some_xmits(conn);
    if (rp==OOP_CONTINUE) {
      loop->on_fd(loop, conn->fd, OOP_WRITE, conn_writeable, conn);
      return;
    } else if (rp==OOP_HALT) {
      return;
    } else if (!rp) {
      /* transmitted everything */
    } else {
      abort();
    }
  }
}


/*========== article transmission ==========*/

static XmitDetails *xmit_core(Conn *conn, const char *data, int len,
                  XmitKind kind) { /* caller must then fill in details */
  struct iovec *v= &conn->xmit[conn->xmitu];
  XmitDetails *d= &conn->xmitd[conn->xmitu++];
  v->iov_base= data;
  v->iov_len= len;
  d->kind= kind;
  return d;
}

static void xmit_noalloc(Conn *conn, const char *data, int len) {
  xmit_core(conn,data,len, xk_Const);
}
#define XMIT_LITERAL(lit) (xmit_noalloc(conn, (lit), sizeof(lit)-1))

static void xmit_artbody(Conn *conn, ARTHANDLE *ah /* consumed */) {
  XmitDetails *d= xmit_core(conn, ah->data, ah->len, sk_Artdata);
  d->info.sm_art= ah;
}

static void xmit_free(XmitDetails *d) {
  switch (d->kind) {
  case xk_Malloc:  free(d->info.malloc_tofree);   break;
  case xk_Artdata: SMfreearticle(d->info.sm_art); break;
  case xk_Const:                                  break;
  default: abort();
  }
}

static void *conn_write_some_xmits(Conn *conn) {
  /* return values:
   *      0:            nothing more to write, no need to call us again
   *      OOP_CONTINUE: more to write but fd not writeable
   *      OOP_HALT:     disaster, have destroyed conn
   */
  for (;;) {
    int count= conn->xmitu;
    if (!count) return 0;

    if (count > IOV_MAX) count= IOV_MAX;
    ssize_t rs= writev(conn->fd, conn->xmit, count);
    if (rs < 0) {
      if (errno == EAGAIN) return OOP_CONTINUE;
      connfail(conn, "write failed: %s", strerror(errno));
      return OOP_HALT;
    }
    assert(rs > 0);

    for (done=0; rs && done<xmitu; done++) {
      struct iovec *vp= &conn->xmit[done];
      XmitDetails *dp= &conn->xmitd[done];
      if (rs > vp->iov_len) {
	rs -= vp->iov_len;
	xmit_free(dp);
      } else {
	vp->iov_base += rs;
	vp->iov_len -= rs;
      }
    }
    int newu= conn->xmitu - done;
    memmove(conn->xmit,  conn->xmit  + done, newu * sizeof(*conn->xmit));
    memmove(conn->xmitd, conn->xmitd + done, newu * sizeof(*conn->xmitd));
    conn->xmitu= newu;
  }
}

static void conn_make_some_xmits(Conn *conn) {
  for (;;) {
    if (conn->xmitu+5 > CONNIOVS)
      break;

    Article *art= LIST_REMHEAD(queue);
    if (!art) break;

    if (art->checked || (conn->stream && nocheck)) {
      /* actually send it */

      ARTHANDLE *artdata= SMretrieve();

      if (conn->stream) {
	if (artdata) {
	  XMIT_LITERAL("TAKETHIS ");
	  xmit_noalloc(conn, art->mid, art->midlen);
	  XMIT_LITERAL("\r\n");
	  xmit_artbody(conn, artdata);
	}
      } else {
	/* we got 235 from IHAVE */
	if (artdata) {
	  xmit_artbody(conn, artdata);
	} else {
	  XMIT_LITERAL(".\r\n");
	}
      }

      art->sent= 1;
      LIST_ADDTAIL(conn->sent, art);

      counts[art->checked].sent++;

    } else {
      /* check it */

      if (conn->stream)
	XMIT_LITERAL("IHAVE ");
      else
	XMIT_LITERAL("CHECK ");
      xmit_noalloc(art->mid, art->midlen);
      XMIT_LITERAL("\r\n");

      LIST_ADDTAIL(conn->sent, art);
      counts[art->checked].offered++;
    }
  }
}


/*========== handling responses from peer ==========*/

static const oop_rd_style peer_rd_style= {
  OOP_RD_DELIM_STRIP, '\n',
  OOP_RD_NUL_FORBID,
  OOP_RD_SHORTREC_FORBID
};

static Article *article_reply_check(Connection *conn, const char *response,
				    int code_indicates_streaming,
				    const char *sanitised_response) {
  Article *art= LIST_REMHEAD(conn->sent);

  if (!art) {
    connfail(conn,
	     "peer gave unexpected response when no commands outstanding: %s",
	     sanitised_response);
    return 0;
  }

  if (code_indicates_streaming) {
    assert(!memchr(response, 0, 4)); /* ensured by peer_rd_ok */
    if (!conn->stream) {
      connfail("peer gave streaming response code "
	       " to IHAVE or subsequent body: %s", sanitised_response);
      return 0;
    }
    const char *got_mid= response+4;
    int got_midlen= strcspn(got_mid, " \n\r");
    if (got_midlen<3 || got_mid[0]!='<' || got_mid[got_midlen-1]!='>') {
      connfail("peer gave streaming response with syntactically invalid"
	       " messageid: %s", sanitised_response);
      return 0;
    }
    if (got_midlen != art->midlen ||
	memcmp(got_mid, art->messageid, got_midlen)) {
      connfail("peer gave streaming response code to wrong article -"
	       " probable synchronisation problem; we offered: %s;"
	       " peer said: %s",
	       art->messageid, sanitised_response);
      return 0;
    }
  } else {
    if (conn->stream) {
      connfail("peer gave non-streaming response code to CHECK/TAKETHIS: %s",
	       sanitised_response);
      return 0;
    }
  }

  return art;
}

static void update_nocheck(int accepted) {
  accept_proportion *= accept_decay;
  accept_proportion += accepted;
  nocheck= accept_proportion >= nocheck_thresh;
  if (nocheck && !nocheck_reported) {
    notice("entering nocheck mode for the first time");
    nocheck_reported= 1;
  }
}

static void article_done(Connection *conn, Article *art, int whichcount) {
  *count++;
  counts.articles[art->checked][whichcount]++;
  if (whichcount == RC_accepted) update_nocheck(1);
  else if (whichcount == RC_unwanted) update_nocheck(0);

  InputFile *ipf= art->ipf;
  while (art->blanklen) {
    static const char spaces[]=
      "                                                                "
      "                                                                "
      "                                                                "
      "                                                                ";
    int w= art->blanklen;  if (w >= sizeof(spaces)) w= sizeof(spaces)-1;
    int r= pwrite(ipf->fd, spaces, w, art->offset);
    if (r==-1) {
      if (errno==EINTR) continue;
      sysdie("failed to blank entry for %s (length %d at offset %lu) in %s",
	     art->messageid, art->blanklen, art->offset, ipf->path);
    }
    assert(r>=0 && r<=w);
    art->blanklen -= w;
    art->offset += w;
  }

  ipf->inprogress--;
  assert(ipf->inprogress >= 0);

  free(art);
}

static void *peer_rd_err(oop_source *lp, oop_read *oread, oop_event ev,
			 const char *errmsg, int errnoval,
			 const char *data, size_t recsz, void *conn_v) {
  Conn *conn= conn_v;
  connfail(conn, "error receiving from peer: %s", errmsg);
  return OOP_CONTINUE;
}

static void *peer_rd_ok(oop_source *lp, oop_read *oread, oop_event ev,
			const char *errmsg, int errnoval,
			const char *data, size_t recsz, void *conn_v) {
  Conn *conn= conn_v;

  if (ev == OOP_RD_EOF) {
    connfail(conn, "unexpected EOF from peer");
    return OOP_CONTINUE;
  }
  assert(ev == OOP_RD_OK);

  char *ep;
  unsigned long code= strtoul(data, &ep, 10);
  if (ep != data+3 || *ep != ' ' || data[0]=='0') {
    char sanibuf[100];
    const char *p= data;
    char *q= sanibuf;
    *q++= '`';
    for (;;) {
      if (q > sanibuf+sizeof(sanibuf)-8) { strcpy(q,"..."); break; }
      int c= *p++;
      if (!c) { *q++= '\''; break; }
      if (c>=' ' && c<=126 && c!='\\') { *q++= c; continue; }
      sprintf(q,"\\x%02x",c);
      q += 4;
    }
    connfail(conn, "badly formatted response from peer: %s", sanibuf);
    return OOP_CONTINUE;
  }

  if (conn->quitting) {
    if (code!=205) {
      connfail(conn, "peer gave failure response to QUIT: %s", sani);
      return OOP_CONTINUE;
    }
    conn close ok;
    return;
  }

  Article *art;

#define GET_ARTICLE							    \
  art= article_reply_check(conn, data, code_streaming, sani);		    \
  if (art) ; else return OOP_CONTINUE /* reply_check has failed the conn */

#define ARTICLE_DEALTWITH(streaming,how)			\
  code_streaming= (streaming)					\
  GET_ARTICLE;							\
  article_done(conn, art, RC_##how);  break;

#define PEERBADMSG(m) connfail(conn, m ": %s", sani);  return OOP_CONTINUE

  int code_streaming= 0;

  switch (code) {

  case 400: PEERBADMSG("peer stopped accepting articles");
  case 503: PEERBADMSG("peer timed us out");
  default:  PEERBADMSG("peer sent unexpected message");

  case 435: ARTICLE_DEALTWITH(0,unwanted); /* IHAVE says they have it */
  case 438: ARTICLE_DEALTWITH(1,unwanted); /* CHECK/TAKETHIS: they have it */

  case 235: ARTICLE_DEALTWITH(0,accepted); /* IHAVE says thanks */
  case 239: ARTICLE_DEALTWITH(1,accepted); /* TAKETHIS says thanks */

  case 437: ARTICLE_DEALTWITH(0,rejected); /* IHAVE says rejected */
  case 439: ARTICLE_DEALTWITH(1,rejected); /* TAKETHIS says rejected */

  case 238: /* CHECK says send it */
    code_streaming= 1;
  case 335: /* IHAVE says send it */
    GET_ARTICLE;
    count_checkedwanted++;
    LIST_ADDTAIL(conn->queue);
    if (art->checked) {
      connfail("peer gave %d response to article body: %s",code, sani);
      return OOP_CONTINUE;
    }
    art->checked= 1;
    break;

  case 431: /* CHECK or TAKETHIS says try later */
    code_streaming= 1;
  case 436: /* IHAVE says try later */
    GET_ARTICLE;
    if (fprintf(defer, "%s %s\n", TokenToText(art->token), art->messageid) <0
	|| fflush(defer))
      sysdie("write to defer file %s",path_ductdefer);
    article_done(conn, art, RC_deferred);
    break;

  }

  check_check_work(conn);
  return OOP_CONTINUE;
}

 
/*========== monitoring of input files ==========*/

static void feedfile_eof(InputFile *ipf) {
  assert(ipf != main_input_file); /* promised by tailing_try_read */
  assert(ipf == old_input_file);
  assert(sms == sm_SEPARATED);
  sms= sm_FINISHING;
  inputfile_tailing_stop(ipf);
  inputfile_tailing_start(main_input_file);
}

static InputFile *open_input_file(const char *path) {
  int fd= open(path, O_RDONLY);
  if (fd<0) {
    if (errno==ENOENT) return 0;
    sysdie("unable to open input file %s", path);
  }

  InputFile *ipf= xmalloc(sizeof(InputFile));
  memset(ipf,0,sizeof(*ipf));
  
  ipf->readable.on_readable= tailing_on_readable;
  ipf->readable.on_cancel=   tailing_on_cancel;
  ipf->readable.try_read=    tailing_try_read;

  ipf->fd= fd;
  ipf->path= path;
  
  return ipf;
}

static void close_input_file(InputFile *ipf) {
  assert(!ipf->readable_callback); /* must have had ->on_cancel */
  assert(!ipf->filemon); /* must have had inputfile_tailing_stop */
  assert(!ipf->rd); /* must have had inputfile_tailing_stop */
  assert(!ipf->inprogress); /* no dangling pointers pointing here */

  if (close(ipf->fd)) sysdie("could not close input file %s", ipf->path);
  free(ipf);
}


/*---------- dealing with articles read in the input file ----------*/

typedef void *feedfile_got_article(oop_source *lp, oop_read *rd,
				   oop_rd_event ev, const char *errmsg,
				   int errnoval,
				   const char *data, size_t recsz,
				   void *ipf_v) {
  InputFile *ipf= ipf_v;
  Article *art;
  char tokentextbuf[sizeof(TOKEN)*2+3];

  if (!data) { feedfile_eof(ipf); return OOP_CONTINUE; }

  if (data[0] && data[0]!=' ') {
    char *space= strchr(data,' ');
    int tokenlen= space-data;
    int midlen= (int)recsz-tokenlen-1;
    if (midlen < 0) goto bad_data;

    if (tokenlen != sizeof(TOKEN)*2+2) goto bad_data;
    memcpy(tokentextbuf, data, tokenlen);
    tokentextbuf[tokenlen]= 0;
    if (!IsToken(tokentextbuf)) goto bad_data;

    art= xmalloc(sizeof(*art) - 1 + midlen + 1);
    art->offset= ipf->offset;
    art->blanklen= recsz;
    art->midlen= midlen;
    art->checked= art->sentbody= 0;
    art->ipf= ipf;  ipf->inprogress++;
    art->token= TextToToken(tokentextbuf);
    strcpy(art->messageid, space+1);
    LIST_ADDTAIL(queue, art);
  }
  ipf->offset += recsz + 1;

  if (sms==sm_NORMAL && ipf->offset >= flush_threshold) {
    int r= link(feedfile, duct_path);
    if (r) sysdie("link feedfile %s to ductfile %s", feedfile, dut_path);
    /* => Hardlinked */

    r= unlink(feedfile);
    if (r) sysdie("unlink old feedfile link %s", feedfile);
    /* => Moved */

    spawn_inndcomm_flush(); /* => Flushing, sets sms to sm_FLUSHING */
  }

  check_master_queue();
}


/*========== tailing input file ==========*/

static void filemon_start(InputFile *ipf) {
  assert(!ipf->filemon);

  ipf->filemon= xmalloc(sizeof(*ipf->filemon));
  memset(ipf->filemon, 0, sizeof(*ipf->filemon));
  filemon_method_startfile(ipf, ipf->filemon);
}

static void filemon_stop(InputFile *ipf) {
  if (!ipf->filemon) return;
  filemon_method_stopfile(ipf, ipf->filemon);
  free(ipf->filemon);
  ipf->filemon= 0;
}

static void filemon_callback(InputFile *ipf) {
  ipf->readable_callback(ipf->readable_callback_user);
}

static void *tailing_rable_call_time(oop_source *loop, struct timeval tv,
				     void *user) {
  InputFile *ipf= user;
  return ipf->readable_callback(ipf->readable_callback_user);
}

static void on_cancel(struct oop_readable *rable) {
  InputFile *ipf= (void*)rable;

  if (ipf->filemon) filemon_stopfile(ipf);
  loop->cancel_time(loop, OOP_TIME_NOW, tailing_rable_call_time, ipf);
  ipf->readable_callback= 0;
}

static int tailing_on_readable(struct oop_readable *rable,
				oop_readable_call *cb, void *user) {
  InputFile *ipf= (void*)rable;

  tailing_on_cancel(rable);
  ipf->readable_callback= cb;
  ipf->readable_callback_user= user;
  filemon_startfile(ipf);

  loop->on_time(loop, OOP_TIME_NOW, tailing_rable_call_time, ipf);
  return 0;
}

static ssize_t tailing_try_read(struct oop_readable *rable, void *buffer,
				size_t length) {
  InputFile *ipf= (void*)rable;
  for (;;) {
    ssize_t r= read(ipf->fd, buffer, length);
    if (!r && ipf==main_input_file) { errno=EAGAIN; return -1; }
    if (r==-1 && errno==EINTR) continue;
    return r;
  }
}

/*---------- filemon implemented with inotify ----------*/

#if defined(HAVE_INOTIFY) && !defined(HAVE_FILEMON)
#define HAVE_FILEMON

#include <linux/inotify.h>

static int filemon_inotify_fd;
static int filemon_inotify_wdmax;
static InputFile **filemon_inotify_wd2ipf;

typedef struct Filemon_Perfile {
  int wd;
} Filemon_Inotify_Perfile;

static void filemon_method_startfile(InputFile *ipf, Filemon_Perfile *pf) {
  int wd= inotify_add_watch(filemon_inotify_fd, ipf->path, IN_MODIFY);
  if (wd < 0) sysdie("inotify_add_watch %s", ipf->path);

  if (wd >= filemon_inotify_wdmax) {
    int newmax= wd+2;
    filemon_inotify_wd= xrealloc(filemon_inotify_wd2ipf,
				 sizeof(*filemon_inotify_wd2ipf) * newmax);
    memset(filemon_inotify_wd2ipf + filemon_inotify_wdmax, 0,
	   sizeof(*filemon_inotify_wd2ipf) * (newmax - filemon_inotify_wdmax));
    filemon_inotify_wdmax= newmax;
  }

  assert(!filemon_inotify_wd2ipf[wd]);
  filemon_inotify_wd2ipf[wd]= ipf;

  pf->wd= wd;
}

static void filemon_method_stopfile(InputFile *ipf, Filemon_Perfile *pf) {
  int wd= pf->wd;
  int r= inotify_rm_watch(filemon_inotify_fd, filemon_inotify_wd);
  if (r) sysdie("inotify_rm_watch");
  filemon_inotify_wd2ipf[wd]= 0;
}

static void *filemon_inotify_readable(oop_source *lp, int fd,
				      oop_event e, void *u) {
  struct inotify_event iev;
  for (;;) {
    int r= read(filemon_inotify_fd, &iev, sizeof(iev));
    if (r==-1) {
      if (errno==EAGAIN) break;
      sysdie("read from inotify master");
    } else if (r==sizeof(iev)) {
      assert(iev.wd >= 0 && iev.wd < filemon_inotify_wdmax);
    } else {
      die("inotify read %d bytes wanted struct of %d", r, (int)sizeof(iev));
    }
    InputFile *ipf= filemon_inotify_wd2ipf[iev.wd];
    filemon_callback(ipf);
  }
  return OOP_CONTINUE;
}

static int filemon_method_init(void) {
  filemon_inotify_fd= inotify_init();
  if (filemon_inotify_fd<0) {
    syswarn("could not initialise inotify: inotify_init failed");
    return 0;
  }
  set nonblock;
  loop->on_fd(loop, filemon_inotify_fd, OOP_READ, filemon_inotify_readable);

  return 1;
}

#endif /* HAVE_INOTIFY && !HAVE_FILEMON *//

/*---------- filemon dummy implementation ----------*/

#if !defined(HAVE_FILEMON)

typedef struct Filemon_Perfile { int dummy; } Filemon_Dummy_Perfile;

static int filemon_method_init(void) { return 0; }
static void filemon_method_startfile(InputFile *ipf, Filemon_Perfile *pf) { }
static void filemon_method_stopfile(InputFile *ipf, Filemon_Perfile *pf) { }

#endif /* !HAVE_FILEMON */

/*---------- interface to start and stop an input file ----------*/

static const oop_rd_style feedfile_rdstyle= {
  OOP_RD_DELIM_STRIP, '\n',
  OOP_RD_NUL_FORBID,
  OOP_RD_SHORTREC_EOF,
};

static void inputfile_tailing_start(InputFile *ipf) {
  assert(!ipf->fd);
  ipf->readable->on_readable= tailing_on_readable;
  ipf->readable->on_cancel=   tailing_on_cancel;
  ipf->readable->try_read=    tailing_try_read;
  ipf->readable->delete_tidy= 0; /* we never call oop_rd_delete_{tidy,kill} */
  ipf->readable->delete_kill= 0;

  ipf->readable_callback= 0;
  ipf->readable_callback_user= 0;

  ipf->rd= oop_rd_new(loop, &ipf->readable, 0,0);
  assert(ipf->fd);

  int r= oop_rd_read(ipf->rd, &feedfile_rdstyle, MAX_LINE_FEEDFILE,
		     feedfile_got_article,ipf, feedfile_problem,ipf);
  if (r) sysdie("unable start reading feedfile %s",ipf->path);
}

static void inputfile_tailing_stop(InputFile *ipf) {
  assert(ipf->fd);
  oop_rd_delete(ipf->rd);
  ipf->rd= 0;
  assert(!ipf->filemon); /* we shouldn't be monitoring it now */
}


/*========== interaction with innd ==========*/

/* See official state diagram at top of file.  We implement
 * this as follows:
 *
	   ================
	    WAITING
	   [Nothing/Noduct]
	    poll for F
	   ================
		|
		|     TIMEOUT
		|`--------------------------.
	   	|                           | install defer as backlog
                | OPEN F SUCCEEDS           | exit
     ,--------->|                           V
     |          V                         =========
     |     ========                        (ESRCH)
     |      NORMAL                        [Dropped]
     |     [Normal]                       =========
     |      read F
     |     ========
     |          |
     |          | F IS SO BIG WE SHOULD FLUSH
     ^          | hardlink F to D
     |     [Hardlinked]
     |          | unlink F
     |          | our handle onto F is now onto D
     |     [Moved]
     |          |
     |          |<---------------------------------------------------.
     |          |                                                    |
     |          | spawn inndcomm flush                               |
     |          V                                                    |
     |     ==========                                                |
     |      FLUSHING                                                 |
     |     [Flushing]                                                |
     |      read D                                                   |
     |     ==========                                                |
     |          |                                                    |
     |          |   INNDCOMM FLUSH FAILS                             ^
     |          |`----------------------->--------.                  |
     |          |                                 |                  |
     |          |   NO SUCH SITE                  V                  |
     ^          |`----------------.            =========             |
     |          |                 |            FLUSHFAIL             |
     |          |                 V            [Moved]               |
     |          |            ==========        read D                |
     |          |             DROPPING         =========             |
     |          |            [Dropping]           |                  |
     |          |             read D              | TIME TO RETRY    |
     |          |            ==========           `------------------'
     |          | FLUSH OK        |
     |          | open F          | AT EOF OF D AND ALL PROCESSED
     |          V                 | install defer as backlog
     |     ===========            | unlink D
     |      SEPARATED             | exit
     |     [Separated]            V
     |      read D            ==========
     |     ===========         (ESRCH)
     |          |             [Droppped]
     |          |             ==========
     |          V
     |          | AT EOF OF D
     ^          |
     |     ===========
     |      FINISHING
     |     [Finishing]
     |      read F
     |      write D
     |     ===========
     |          |
     |          | ALL D PROCESSED
     |          | install defer as backlog
     |          | start new defer
     ^          V unlink D
     |          | close D
     |          |
     `----------'

 *
 */

static void open_defer(void) {
  struct stat stab;

  assert(!defer);
  defer= fopen(path_ductdefer, "a+");
  if (!defer) sysdie("could not open defer file %s", path_ductdefer);

  /* truncate away any half-written records */

  r= fstat(fileno(defer), &stab);
  if (r) sysdie("could not stat newly opened defer file %s", path_ductdefer);

  if (stab.st_size > LONG_MAX)
    die("defer file %s size is far too large", path_ductdefer);

  if (!stab.st_size)
    return;

  long orgsize= stab.st_size;
  long truncto= stab.st_size;
  for (;;) {
    if (!truncto) break; /* was only (if anything) one half-truncated record */
    if (fseek(defer, truncto-1, SEEK_SET) < 0)
      sysdie("seek in defer file %s while truncating partial", path_ductdefer);

    r= getc(defer);
    if (r==EOF) {
      if (ferror(defer))
	sysdie("failed read from defer file %s", path_ductdefer);
      else
	die("defer file %s shrank while we were checking it!", path_ductdefer);
    }
    if (r=='\n') break;
    truncto--;
  }

  if (stab.st_size != truncto) {
    warn("truncating half-record at end of defer file %s -"
	 " shrinking by %ld bytes from %ld to %ld",
	 path_ductdefer, orgsize - truncto, orgsize, truncto);

    if (fflush(defer)) sysdie("could not flush defer file %s", path_ductdefer);
    if (ftruncate(fileno(defer), truncto))
      sysdie("could not truncate defer file %s", path_ductdefer);

  } else {
    info("continuing existing defer file %s (%ld bytes)",
	 path_ductdefer, orgsize);
  }
  if (fseek(defer, truncto, SEEK_SET))
    sysdie("could not seek to new end of defer file %s", path_ductdefer);
}

static void statemc_init(void) {
  struct stat stab;

  path_ductlock=  xasprintf("%s_duct.lock",  feedfile);
  path_duct=      xasprintf("%s_duct",       feedfile);
  path_ductdefer= xasprintf("%s_duct.defer", feedfile);

  if (lstat(path_ductdefer, &stab)) {
    if (errno!=ENOENT) sysdie("could not check defer file %s", path_defer);
  } else {
    if (!S_ISREG(stab.st_mode))
      die("defer file %s not a plain file (mode 0%lo)",
	  path_defer, (unsigned long)stab.st_mode);
    switch (stab.st_nlink==1) {
    case 1: /* ok */ break;
    case 2:
      if (unlink(path_defer))
	sysdie("could not unlink stale defer file link %s (presumably"
	       " hardlink to backlog file)", path_defer);
      break;
    default:
      die("defer file %s has unexpected link count %d",
	  path_defer, stab.st_nlink);
    }
  }
  open_defer();

  int lockfd= open(path_ductlock, O_CREAT|O_RDWR, 0600);
  if (lockfd<0) sysdie("open lockfile %s", path_ductlock);

  struct flock fl;
  memset(&fl,0,sizeof(fl));
  fl.l_type= F_WRLCK;
  fl.l_whence= SEEK_SET;
  r= fcntl(lockfd, F_SETLK, &fl);
  if (r==-1) {
    if (errno==EACCES || errno==EAGAIN)
      die("another duct holds the lockfile");
    sysdie("fcntl F_SETLK lockfile %s", path_ductlock);
  }

  InputFile *file_d= open_input_file(path_duct);

  if (file_d) {
    struct stat stab_f, stab_d;

    r= stat(feedfile, &stab_f);
    if (r) {
      if (errno!=ENOENT) sysdie("check feed file %s", feedfile);
      /* D exists, F ENOENT => Moved */
      goto found_moved;
    }

    /* F and D both exist */

    r= fstat(file_d->fd, &stab_d);
    if (r) sysdie("check duct file %s", ductfile);

    if (stab_d.st_ino == stab_f.st_ino &&
	stab_d.st_dev == stab_f.st_dev) {
      /* F==D => Hardlinked*/
      r= unlink(path_duct);
      if (r) sysdie("unlink feed file %s during startup", feedfile);
    found_moved:
      /* => Moved */
      startup_set_input_file(file_d);
      spawn_inndcomm_flush(); /* => Flushing, sets sms to sm_FLUSHING */
    } else {
      /* F!=D => Separated */
      sms= sm_SEPARATED;
      startup_set_input_file(file_d);
    }
  } else { /*!file_d*/
    sms= sm_WAITING;
    statemc_waiting_poll();
  }
}

static void statemc_poll(void) {
  if (sms == sm_WAITING) statemc_waiting_poll();
  if (sms == sm_FINISHING && !old_input_file->inprogress) statemc_finishdone();
}

static void statemc_waiting_poll(void) {
  InputFile *file_f= open_input_file(feedfile);
  if (!file_f) {
    if (waiting_periods_sofar++ > waiting_timeout_periods)
      die("timed out waiting for innd to create feed file %s", feedfile);
    return;
  }
  startup_set_input_file(file_d);
  sms= sm_NORMAL;
}

static void startup_set_input_file(InputFile *f) {
  assert(!main_input_file);
  main_input_file= f;
  inputfile_tailing_start(f);
}

static void statmc_finishdone(void) {
  time_t now;
  struct stat stab;

  assert(sms == sm_FINISHING);

  r= fstat(fileno(defer), &stab);
  if (r) sysdie("check defer file %s", path_defer);

  if (fclose(defer)) sysdie("could not close defer file %s", path_defer);
  defer= 0;

  now= time(0);
  if (now==-1) sysdie("could not get current time for backlog filename");

  char *backlog= xasprintf("%s_backlog_%lu.%lu", feedfile,
			   (unsigned long)now,
			   (unsigned long)stab.st_ino);
  if (link(path_defer, path_backlog))
    sysdie("could not install defer file %s as backlog file %s",
	   path_defer, backlog);
  if (unlink(path_defer))
    sysdie("could not unlink old defer link %s to backlog file %s",
	   path_defer, backlog);
  open_defer();

  close_input_file(old_input_file);
  old_input_file= 0;

  if (unlink(path_duct))
    sysdie("could not unlink old duct file %s", path_duct);

  sms= sm_NORMAL;
}

/*========== flushing the feed ==========*/

 

/*========== main program ==========*/

static void postfork_inputfile(InputFile *ipf) {
  if (!ipf) return;
  assert(ipf->fd >= 0);
  close(ipf->fd);
  ipf->fd= -1;
}

static void postfork_conns(Connection *conn) {
  while (conn) {
    close(conn->fd);
    conn= conn->next;
  }
}

static void postfork_stdio(FILE *f) {
  /* we have no stdio streams that are buffered long-term */
  if (f) fclose(f);
}

static void postfork(const char *what) {
  if (signal(SIGPIPE, SIG_DFL) == SIG_ERR)
    sysdie("%s child: failed to reset SIGPIPE");

  postfork_inputfile(main_input_file);
  postfork_inputfile(old_input_file);
  postfork_conns(idle.head);
  postfork_conns(working.head);
  postfork_conns(full.head);
  postfork_stdio(defer);
}


#define EVERY(what, interval, body)					     \
  static const struct timeval what##_timeout = { 5, 0 };		     \
  static void what##_schedule(void);					     \
  static void *what##_timedout(oop_source *lp, struct timeval tv, void *u) { \
    { body }								     \
    what##_schedule();							     \
  }									     \
  static void what##_schedule(void) {					     \
    loop->on_time(loop, what##_timeout, what##_timedout, 0);		     \
  }

EVERY(filepoll, {5,0}, { check_master_queue(); })

EVERY(period, {PERIOD_SECONDS,0}, {
  if (connect_delay) connect_delay--;
  statemc_poll();
  check_master_queue();
});

main {
  ignore sigpipe;
  if (!filemon_init())
    filepoll_schedule();
  period_schedule();
};
